 '--------------------------------------------------
 '
 '  Title   :  Simple Expression Evaluator
 '  Version :  1.01
 '  Author  :  Aleksandar Ruzicic (admin@krcko.net)  
 '  File    :  fbeval.bas
 '
 '
 ' BIG thanks goes to Jack W. Crenshaw for his
 ' "LET'S BUILD A COMPILER!" text series.
 '
 ' You're free to use this in any way you find it
 ' useful, just "give credit where credit is due" :)
 ' --------------------------------------------------

#ifndef __FBEVAL_BAS__
#define __FBEVAL_BAS__

Namespace Eval

Enum Boolean
  False=0
  True=Not False ' equals to -1
End Enum

Enum EvalErrorMessage
    ErrorSuccess = 0    ' no error
    ErrorToken          ' unexpected token
    ErrorBrace          ' unbalanced braces
    ErrorDivByZero      ' division by zero (not implemented...)
    ErrorUnknownVar     ' undefined variable
    ErrorUnknownFunc    ' undefined function
End Enum

Type EvalError
    Number      As EvalErrorMessage
    Description As String
    Position    As Integer
    Tag         As String
End Type

Type Result    
    Value       As Double
    Success     As Boolean
    Error       As EvalError
End Type

Type EvalVariable
    Name        As String
    Value       As Double
End Type

Type EvalFunction
    Name        As String
    Wrapper     As Function(argc As Integer, argv() As Double) As Double
End Type


Declare Sub SetVariable(ByVal VarName As String, Value As Double)
Declare Sub UnsetVariable(ByVal VarName As String)
Declare Sub ClearVariables()
Declare Function GetVariable(VarName As String) As Double

Declare Sub SetFunction(ByVal FuncName As String, Wrapper As Function(argc As Integer, argv() As Double) As Double)
Declare Sub UnsetFunction(ByVal FuncName As String)
Declare Sub ClearFunctions()
Declare Function GetFunction(FuncName As String, argc As Integer, argv() As Double, startPos As Integer) As Double

Declare Sub GetChar()
Declare Sub SkipWhite()

Declare Function IsAlpha(char As String) As Boolean
Declare Function IsDigit(char As String) As Boolean
Declare Function IsHexChar(char As String) As Boolean
Declare Function IsBinChar(char As String) As Boolean
Declare Function IsOctChar(char As String) As Boolean
Declare Function IsAlNum(char As String) As Boolean
Declare Function IsAddop(char As String) As Boolean
Declare Function IsMulop(char As String) As Boolean
Declare Function IsShiftop(char As String) As Boolean
Declare Function IsWhite(char As String) As Boolean

Declare Sub Unexpected()
Declare Function Match(what As String) As Boolean
Declare Sub makeError(ByRef er As Result)

Declare Function GetName() As String
Declare Function GetNum() As Double

Declare Function Identifier() As Double
Declare Function Factor() As Double
Declare Function Expression() As Double
Declare Function AddOp() As Double
Declare Function ShiftOp() As Double
Declare Function ModOp() As Double
Declare Function DivOp() As Double
Declare Function MulOp() As Double
Declare Function UnOp() As Double
Declare Function ExpOp() As Double
Declare Function ImpOp() As Double
Declare Function EqvOp() As Double
Declare Function XorOp() As Double
Declare Function OrOp() As Double
Declare Function AndOp() As Double
Declare Function NotOp() As Double

Declare Function Execute(ExpressionString As String) As Result

Dim Shared As Integer position
Dim Shared As String source, look
Dim Shared As Boolean parsingError
Dim Shared As EvalErrorMessage e
Dim Shared As String errorTag, currentFunction
Dim Shared As EvalVariable variables()
Dim Shared As EvalFunction functions()

Sub SetVariable(ByVal VarName As String, Value As Double)
    
    Dim As Integer i, upper
    
    VarName = UCase$(VarName)
    
    upper = UBound(variables)
    
    For i = 1 To upper
        If variables(i).Name = VarName Then
            variables(i).Value = Value
            Exit Sub
        End If
    Next
    
    i = upper + 1
    
    ReDim Preserve variables(i)
    
    variables(i).Name = VarName
    variables(i).Value = Value
    
End Sub

Sub UnsetVariable(ByVal VarName As String)
    
    Dim As Integer i, k, upper
    
    VarName = UCase$(VarName)
    
    upper = UBound(variables)
    
    For i = 1 To upper
        If variables(i).Name = VarName Then
            For k = i To upper - 1
                variables(k).Name = variables(k + 1).Name
                variables(k).Value = variables(k + 1).Value
            Next
            ReDim Preserve variables(upper - 1)
            Exit Sub
        End If
    Next
    
End Sub

Sub ClearVariables()
    Redim variables(0)
End Sub

Function GetVariable(VarName As String) As Double
    
    If parsingError Then Return 0
    
    Dim As Integer i, upper
    
    upper = UBound(variables)
    
    For i = 1 To upper
        If variables(i).Name = VarName Then
            Return variables(i).Value
        End If
    Next
    
    e = ErrorUnknownVar
    errorTag = VarName
    position -= Len(VarName)
    parsingError = True
    
    Return 0
    
End Function

Sub SetFunction(ByVal FuncName As String, Wrapper As Function(argc As Integer, argv() As Double) As Double)
    
    Dim As Integer i, upper
    
    FuncName = UCase$(FuncName)
    
    upper = UBound(functions)
    
    For i = 1 To upper
        If functions(i).Name = FuncName Then
            functions(i).Wrapper = Wrapper
            Exit Sub
        End If
    Next
    
    i = upper + 1
    
    Redim Preserve functions(i)
    
    functions(i).Name = FuncName
    functions(i).Wrapper = Wrapper
    
End Sub


Sub UnsetFunction(ByVal FuncName As String)
    
    Dim As Integer i, k, upper
    
    FuncName = UCase$(FuncName)
    
    upper = UBound(functions)
    
    For i = 1 To upper
        If functions(i).Name = FuncName Then
            For k = i To upper - 1
                functions(k).Name = functions(k + 1).Name
                functions(k).Wrapper = functions(k + 1).Wrapper
            Next
            ReDim Preserve functions(upper - 1)
            Exit Sub
        End If        
    Next
    
End Sub

Sub ClearFunctions()
    ReDim functions(0)
End Sub

Function GetFunction(FuncName As String, argc As Integer, argv() As Double, startPos As Integer) As Double
    
    If parsingError Then Return 0
    
    Dim As Integer i, upper
    Dim As Double Result
    
    upper = UBound(functions)
    
    For i = 1 To upper
        If functions(i).Name = FuncName Then
            currentFunction = FuncName
            Result = functions(i).Wrapper(argc, argv())
            currentFunction = ""
            Return Result
        End If
    Next
    
    e = ErrorUnknownFunc
    errorTag = FuncName
    position = startPos
    parsingError = True
    
    Return 0
    
End Function


Sub GetChar()
  If parsingError Then Exit Sub
  position += 1
  look = Mid$(source, position, 1)
  If Len(look) = 0 Then ' end of expression
    e = ErrorSuccess
  End If
End Sub

#define upperA 65
#define upperZ 90
#define underscore 95
Function IsAlpha(char As String) As Boolean
    
    Dim As Integer c = Asc(char)
    
    If  ((c >= upperA) And (c <= upperZ)) Or c = underscore Then
        Return True
    End If
    
    Return False
    
End Function


#define digit0 48
#define digit9 57
Function IsDigit(char As String) As Boolean
  Return Instr("0123456789", Char)>0
End Function

#define upperF 70
Function IsHexChar(char As String) As Boolean
  Return Instr("0123456789BCDEF", Char)>0
End Function

Function IsBinChar(char As String) As Boolean
  Return Instr("01", Char)>0
End Function

#define digit7 55
Function IsOctChar(char As String) As Boolean
  Return Instr("01234567", Char)>0
End Function


Function IsAlNum(char As String) As Boolean
  Return (IsAlpha(char) Or IsDigit(char))
End Function


Function IsAddop(char As String) As Boolean
    Return InStr("+-", char)>0
End Function


Function IsMulop(char As String) As Boolean
    
    Return InStr("*/", char)>0

End Function

Function IsShiftop(char As String) As Boolean
  Return InStr("<>", char)>0
End Function


Function IsWhite(char As String) As Boolean
  Return InStr(" "+Chr(9), char)> 0
End Function

Sub SkipWhite()
    
    While IsWhite(look)
        GetChar()
    Wend
    
End Sub


Sub Unexpected()
    
    e = ErrorToken
    errorTag = look
    parsingError = True
    
End Sub


Function Match(what As String) As Boolean
    
    If parsingError Then Return False
    
    If look = what Then
        GetChar()
        SkipWhite()
        Return True
    Else
        Unexpected()
        Return False
    End If

End Function

Function GetName() As String
    
    If parsingError Then Return ""
    
    If Not IsAlpha(look) Then
        Unexpected()
        Return ""
    End If
    
    Dim As String token = ""
    
    While IsAlNum(look)
        token += look
        GetChar()
    Wend
    
    SkipWhite()
    
    Return token
        
End Function


Function GetNum() As Double
    
    If parsingError Then Return 0
    
    Dim As String num = ""
    Dim As Boolean dot = False, isHex = False, isOct = False, isBin = False
    
    If look = "&" Then
        Match("&")
        Select Case look
            Case "H"
                isHex = True
                num = "&H"
                
            Case "B"
                isBin = True
                num = "&B"
            
            Case "O"
                isOct = True
                num = "&O"
            
            Case Else
                Unexpected()
                Return 0
        End Select
        GetChar()
    ElseIf Not (IsDigit(look) Or (look = ".")) Then
        Unexpected()
        Return 0
    End If
    
    If isHex Or isBin Or isOct Then
        While (IsHexChar(look) And isHex) Or (IsBinChar(look) And IsBin) Or (IsOctChar(look) And isOct)
            num += look
            GetChar()
        Wend
    Else
        num = "0"
        While (IsDigit(look) Or (look = "."))
            If look = "." Then
                If dot Then
                    Unexpected()
                    Return 0
                Else
                    num += "."
                    dot = True
                End If
            Else
                num += look
            End If
            GetChar()
        Wend
    End If
    
    SkipWhite()
    
    Return Val(num)
    
End Function

Function Identifier() As Double
    
    If parsingError Then Return 0
    
    Dim As String Ident
    Dim As Double Result, argv()
    Dim As Integer argc, funcPos
    
    funcPos = position
    
    Ident = GetName()
    
    If look = "(" Then
        Match("(")
        
        If look = ")" Then
            Match(")")
            Result = GetFunction(Ident, 0, argv(), funcPos)
        Else
            
            Dim done As Boolean = False
            
            Do
                If parsingError Then Return 0
				
				SkipWhite()                

                argc += 1
                ReDim Preserve argv(argc - 1)
                
                argv(argc - 1) = Expression()
                
                done = (look <> ",")
                
                If Not done Then GetChar() ' eat comma
                
            Loop Until done
            
            If Match(")") Then
                Result = GetFunction(Ident, argc, argv(), funcPos)
            End If
        End If
    Else
        Result = GetVariable(Ident)
    End If
    
    SkipWhite()
    
    Return Result
    
End Function

Function Factor() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    If look = "(" Then
        Match("(")
        Result = Expression()
        If Not Match(")") Then
            e = ErrorBrace
            parsingError = True
        End If
    ElseIf IsAlpha(look) Then
        Result = Identifier()
    Else
        Result = GetNum()
    End If
   
    Return Result
    
End Function

Function Expression() As Double
    Return AddOp()
End Function

Function AddOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = ShiftOp()
    
    While IsAddop(look)
        If parsingError Then Return 0
        Select Case look
            Case "+"
                Match("+")
                Result = Result + ShiftOp()
                
            Case "-"
                Match("-")
                Result = Result - ShiftOp()
        End Select
    Wend
    
    Return Result
    
End Function


Function ShiftOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = ModOp()
    
    While IsShiftop(look)
        If parsingError Then Return 0
        Select Case look
            Case "<"
                Match("<")
                Result = Result Shl ModOp()
                
            Case ">"
                Match(">")
                Result = Result Shr ModOp()
        End Select
    Wend
    
    Return Result
    
End Function


Function ModOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = DivOp()
    
    While look = "%"
        If parsingError Then Return 0
        Match("%")
        Result = Result Mod DivOp()
    Wend
    
    Return Result
    
End Function


Function DivOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = MulOp()
    
    While look = "\"
        If parsingError Then Return 0
        Match("\")
        Result = Result \ MulOp()
    Wend
    
    Return Result
    
End Function

Function MulOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result, divider
    
    Result = UnOp()
    
    While IsMulop(look)
        If parsingError Then Return 0
        Select Case look
            Case "*"
                Match("*")
                Result = Result * UnOp()
                
            Case "/"
                Match("/")
                Result = Result / UnOp()
        End Select
    Wend
    
    Return Result
    
End Function

Function UnOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    If IsAddop(look) Then
        Select Case look
            Case "+"
                Match("+")
                Result = ExpOp()
                
            Case "-"
                Match("-")
                Result = -ExpOp()
        End Select
    Else
        Result = ExpOp()
    End If
    
    Return Result
    
End Function

Function ExpOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = ImpOp()
    
    While look = "^"
        If parsingError Then Return 0
        Match("^")
        Result = Result ^ ImpOp()
    Wend
    
    Return Result
    
End Function

Function ImpOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = EqvOp()
    
    While look = "@"
        If parsingError Then Return 0
        Match("@")
        Result = Result Imp EqvOp()
    Wend
    
    Return Result
    
End Function

Function EqvOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = XorOp()
    
    While look = "="
        If parsingError Then Return 0
        Match("=")
        Result = Result Eqv XorOp()
    Wend
    
    Return Result
    
End Function

Function XorOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = OrOp()
    
    While look = "$"
        If parsingError Then Return 0
        Match("$")
        Result = Result Xor OrOp()
    Wend
    
    Return Result
    
End Function

Function OrOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = AndOp()
    
    While look = "|"
        If parsingError Then Return 0
        Match("|")
        Result = Result Or AndOp()
    Wend
    
    Return Result
    
End Function

Function AndOp() As Double
    
    If parsingError Then Return 0
    
    Dim As Double Result
    
    Result = NotOp()
    
    While look = "&"
        If parsingError Then Return 0
        Match("&")
        Result = Result And NotOp()
    Wend
    
    Return Result
    
End Function

Function NotOp() As Double
  If parsingError Then Return 0
  Dim As Double Result
  If look = "!" Then
    Result = Not Factor()
  Else
    Result = Factor()
  End If
  Return Result
End Function
    
Function Execute(ExpressionString As String) As Result
  Dim As Result eResult
  source = UCase$(ExpressionString)
  position = 0
  e = ErrorSuccess
  parsingError = False
  errorTag = ""
  GetChar()
  SkipWhite()
  eResult.Value = Expression()
  makeError(eResult)
  Return eResult
End Function


Sub makeError(ByRef er As Result)
  Dim eMessage(5) As String
  eMessage(0) = "No error"
  eMessage(1) = "Unexpected token"
  eMessage(2) = "Unbalanced braces"
  eMessage(3) = "Division by zero"
  eMessage(4) = "Undefined variable"
  eMessage(5) = "Undefined function"
  er.Success = (e = ErrorSuccess)
  er.Error.Number = e
  er.Error.Description = eMessage(e)
  er.Error.Position = position
  er.Error.Tag = errorTag
End Sub

End Namespace

#endif